# -*- coding: utf-8 -*-

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

## MPHYS interface
# Adrien Crovato
#
# Structure:
#   - AeroBuilder
#       - AeroSurfaceMesh
#       - AeroGroup
#           - Morpher
#           - Solver
#           - Loads
#       - AeroCoefficients
#
# Remarks:
# - for each component that uses the openMDAO matrix-free API, the partial
#   gradients are first computed internally in dart::Adjoint, through
#   om.ImplicitComponent.linearize() or om.ExplicitComponent.compute_partials(),
#   and the matrix-vector product is then computed in dart:: Adjoint through
#   om.ImplicitComponent.apply_linear() or om.ExplicitComponent.compute_jacvec_product().
#   This allows faster gradient evaluations, at the cost of a reasonable
#   increase in memory storage, without interfacing any sparse matrices through
#   python.

import numpy as np
import openmdao.api as om
from mphys.builder import Builder

# Surface mesh
class DartMesh(om.IndepVarComp):
    """Initial surface mesh of moving body
    For compatibility, the node coordinates of connected surfaces are always 3D
    For compatibility, the output is distributed, though it should always be serial

    Attributes
    ----------
    dim : int
        Problem dimension
    bnd : dart.Body object
        Moving body
    """
    def initialize(self):
        self.options.declare('dim', desc='problem dimension')
        self.options.declare('bnd', desc='moving body', recordable=False)

    def setup(self):
        self.bnd = self.options['bnd']
        # Store all the node coordinates and add as output
        x_aero0 = np.zeros(3 * len(self.bnd.nodes)) # surface node coordinates are 3D
        for i in range(len(self.bnd.nodes)):
            for j in range(3):
                x_aero0[3 * i + j] = self.bnd.nodes[i].pos[j]
        self.add_output('x_aero0', distributed=True, val=x_aero0, desc='initial aerodynamic surface node coordinates', tags=['mphys_coordinates'])

    def get_triangulated_surface(self):
        """Triangulate the surface
        Only for 2D surfaces in 3D problems
        Return a list containing a point and two vectors for each surface element
             o p0
            / \
        v0 /___\ v1
        """
        if self.options['dim'] == 2:
            raise RuntimeError('DartMesh - cannot triangulate 1D surface!\n')
        # Create list of point coordinates and vector components
        p0 = [[0.]*3 for _ in range(len(self.bnd.groups[0].tag.elems))]
        v0 = [[0.]*3 for _ in range(len(self.bnd.groups[0].tag.elems))]
        v1 = [[0.]*3 for _ in range(len(self.bnd.groups[0].tag.elems))]
        for i in range(len(self.bnd.groups[0].tag.elems)):
            e = self.bnd.groups[0].tag.elems[i]
            p0i = e.nodes[0].pos
            v0i = e.nodes[1].pos - e.nodes[0].pos
            v1i = e.nodes[2].pos - e.nodes[0].pos
            for j in range(3):
                p0[i][j] = p0i[j]
                v0[i][j] = v0i[j]
                v1[i][j] = v1i[j]
        return [p0, v0, v1]

# Mesh morphers
class DartMorpher(om.ImplicitComponent):
    """Volume mesh morpher
    For compatibility, the node coordinates of connected surfaces are always 3D, while the volume mesh coordinates (only used internaly) can be 2D or 3D
    For compatibility, the input is distributed, though it should always be serial

    Attributes
    ----------
    dim : int
        Problem dimension, used to set the length of the volume node coordinates vector
    bnd : dart.Body object
        Moving body
    mrf : tbox.MshDeform object
        Mesh morpher
    adj : dart.Adjoint object
        Adjoint solver
    """
    def initialize(self):
        self.options.declare('dim', desc='problem dimension')
        self.options.declare('bnd', desc='moving body', recordable=False)
        self.options.declare('mrf', desc='mesh morpher', recordable=False)
        self.options.declare('adj', desc='adjoint solver', recordable=False)

    def setup(self):
        self.dim = self.options['dim']
        self.bnd = self.options['bnd']
        self.mrf = self.options['mrf']
        self.adj = self.options['adj']
        # I/O
        self.add_input('x_aero', distributed=True, shape_by_conn=True, desc='aerodynamic surface node coordinates', tags=['mphys_coupling'])
        self.add_output('xv', val=np.zeros(self.dim * len(self.mrf.msh.nodes)), desc='aerodynamic volume node coordinates', tags=['mphys_coupling']) # volume node coordinates can be 2D or 3D
        # Partials
        # self.declare_partials(of=['xv'], wrt=['x_aero'])

    def solve_nonlinear(self, inputs, outputs):
        """Deform the volume mesh
        """
        # Update inputs
        self.mrf.savePos()
        for i in range(len(self.bnd.nodes)):
            for j in range(3):
                self.bnd.nodes[i].pos[j] = inputs['x_aero'][3 * i + j]
        # Compute
        self.mrf.deform()
        # Update outputs
        for i in range(len(self.mrf.msh.nodes)):
            for j in range(self.dim):
                outputs['xv'][self.dim * i + j] = self.mrf.msh.nodes[i].pos[j]

    def apply_nonlinear(self, inputs, outputs, residuals):
        """Compute the residuals of the mesh coordinates
        """
        raise NotImplementedError('DartMorpher - cannot compute the residuals!\n')

    def linearize(self, inputs, outputs, partials):
        """Compute the mesh jacobian matrix
        """
        self.adj.linearizeMesh()

    def apply_linear(self, inputs, outputs, d_inputs, d_outputs, d_residuals, mode):
        """Perform the matrix-vector product using the mesh Jacobian
        """
        if mode == 'rev':
            if 'xv' in d_residuals:
                if 'xv' in d_outputs:
                    d_outputs['xv'] += self.adj.computeJacVecMesh(d_residuals['xv']) # dX = dRx_dX^T * dRx
                if 'x_aero' in d_inputs:
                    for i in range(len(self.bnd.nodes)):
                        for j in range(self.dim):
                            d_inputs['x_aero'][3 * i + j] -= d_residuals['xv'][self.dim * self.bnd.nodes[i].row + j] # dXs = -dX
        elif mode == 'fwd':
            raise NotImplementedError('DartMorpher - forward mode not implemented!\n')

    def solve_linear(self, d_outputs, d_residuals, mode):
        """Solve for the mesh residuals using the mesh Jacobian
        """
        if mode == 'rev':
            d_residuals['xv'] = self.adj.solveMesh(d_outputs['xv']) # dRx = dRx_dX^-T * dX
        elif mode == 'fwd':
            raise NotImplementedError('DartMorpher - forward mode not implemented!\n')

class DartDummyMorpher(om.ExplicitComponent):
    """Dummy volume mesh morpher
    Allow to define and link the surface mesh coordinates to the volume mesh coordinates, in case the mesh do not deform.
    For compatibility, the node coordinates of connected surfaces are always 3D, while the volume mesh coordinates (only used internaly) can be 2D or 3D
    For compatibility, the input is distributed, though it should always be serial
    """
    def initialize(self):
        self.options.declare('dim', desc='problem dimension')
        self.options.declare('msh', desc='mesh data structure', recordable=False)

    def setup(self):
        dim = self.options['dim']
        msh = self.options['msh']
        # Set initial values
        xv = np.zeros(dim * len(msh.nodes)) # volume node coordinates can be 2D or 3D
        for i in range(len(msh.nodes)):
            for j in range(dim):
                xv[dim * i + j] = msh.nodes[i].pos[j]
        # I/O
        self.add_input('x_aero', distributed=True, shape_by_conn=True, desc='aerodynamic surface node coordinates', tags=['mphys_coupling'])
        self.add_output('xv', val=xv, desc='aerodynamic volume node coordinates', tags=['mphys_coupling'])

    def compute(self, inputs, outputs):
        """DO NOT deform the volume mesh
        """
        pass

    def compute_partials(self, inputs, partials):
        """DO NOT compute the partials derivatives
        """
        pass

# Solver
class DartSolver(om.ImplicitComponent):
    """Aerodynamic solver

    Attributes
    ----------
    sol : dart.Newton object
        Direct Newton solver
    adj : dart.Adjoint object
        Adjoint solver
    """
    def initialize(self):
        self.options.declare('sol', desc='direct solver', recordable=False)
        self.options.declare('adj', desc='adjoint solver', recordable=False)
        self.options.declare('rerr', desc='raise AnalysisError when solver not fully converged')

    def setup(self):
        self.sol = self.options['sol']
        self.adj = self.options['adj']
        # I/O
        self.add_input('aoa', val=0., units='rad', desc='angle of attack', tags=['mphys_input'])
        self.add_input('xv', shape_by_conn=True, desc='aerodynamic volume node coordinates', tags=['mphys_coupling'])
        self.add_output('phi', val=np.zeros(len(self.sol.pbl.msh.nodes)), desc='dart variables (potential)', tags=['mphys_coupling'])
        # Partials
        # self.declare_partials(of=['phi'], wrt=['aoa', 'xv', 'phi'])

    def solve_nonlinear(self, inputs, outputs):
        """Compute the flow variables
        """
        # Update inputs
        # NB: the mesh is already up-to-date mesh already up-to-date because DartMorpher has already been run
        self.sol.pbl.update(inputs['aoa'][0])
        # Compute
        self.sol.reset() # resets ICs (slows down convergence, but increases robustness for aero-structural)
        status = self.sol.run()
        if (status == 1 and self.options['rerr']) or status == 2: # max number of iterations exceeded or NaN in solution
            raise om.AnalysisError('DART solver failed to fully converge!\n')
        # Update outputs
        outputs['phi'] = self.sol.phi

    def apply_nonlinear(self, inputs, outputs, residuals):
        """Compute the residuals of the flow variables
        """
        raise NotImplementedError('DartSolver - cannot compute the residuals!\n')

    def linearize(self, inputs, outputs, partials):
        """Compute the flow jacobian matrix (and the other partials of the flow residuals)
        """
        self.adj.linearizeFlow()

    def apply_linear(self, inputs, outputs, d_inputs, d_outputs, d_residuals, mode):
        """Perform the matrix-vector product using the partial gradients of the flow residuals
        """
        if mode == 'rev':
            if 'phi' in d_residuals:
                if 'phi' in d_outputs:
                    d_outputs['phi'] += self.adj.computeJacVecFlow(d_residuals['phi']) # dU = dRu_dU^T * dRu
                if 'aoa' in d_inputs:
                    d_inputs['aoa'] += self.adj.computeJacVecFlowAoa(d_residuals['phi']) # dA = dRu_dA^T * dRu
                if 'xv' in d_inputs:
                    d_inputs['xv'] += self.adj.computeJacVecFlowMesh(d_residuals['phi']) # dX = dRu_dX^T * dRu
        elif mode == 'fwd':
            raise NotImplementedError('DartSolver - forward mode not implemented!\n')

    def solve_linear(self, d_outputs, d_residuals, mode):
        """Solve for the flow residuals using the flow Jacobian
        """
        if mode == 'rev':
            d_residuals['phi'] = self.adj.solveFlow(d_outputs['phi']) # dRu = dRu_dU^-T * dU
        elif mode == 'fwd':
            raise NotImplementedError('DartSolver - forward mode not implemented!\n')

# Aerodynamic loads
class DartLoads(om.ExplicitComponent):
    """Aerodynamic loads on moving body
    For compatibility, the forces of connected surfaces are always 3D
    For compatibility, the output is distributed, though it should always be serial

    Attributes
    ----------
    qinf : dart.Newton object
        Freestream dynamic pressure
    bnd : dart.Body object
        Moving body
    adj : dart.Adjoint object
        Adjoint solver
    """
    def initialize(self):
        self.options.declare('qinf', desc='freestream dynamic pressure')
        self.options.declare('bnd', desc='moving body', recordable=False)
        self.options.declare('adj', desc='adjoint solver', recordable=False)

    def setup(self):
        self.qinf = self.options['qinf']
        self.bnd = self.options['bnd']
        self.adj = self.options['adj']
        # I/O
        self.add_input('xv', shape_by_conn=True, desc='aerodynamic volume node coordinates', tags=['mphys_coupling'])
        self.add_input('phi', shape_by_conn=True, desc='flow variables (potential)', tags=['mphys_coupling'])
        self.add_output('f_aero', distributed=True, val=np.zeros(3 * len(self.bnd.nodes)), desc='aerodynamic loads', tags=['mphys_coupling'])
        # Partials
        # self.declare_partials(of=['f_aero'], wrt=['xv', 'phi'])

    def compute(self, inputs, outputs):
        """Get the forces on moving body
        """
        # NB: inputs already up-to-date because DartSolver has already been run
        for i in range(len(self.bnd.nodes)):
            for j in range(3):
                outputs['f_aero'][3 * i + j] = self.qinf * self.bnd.nLoads[i][j]

    def compute_partials(self, inputs, partials):
        """Compute the partials gradients of the loads
        """
        self.adj.linearizeLoads(self.bnd)

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        """Perform the matrix-vector product using the partial gradients of the loads
        """
        if mode == 'rev':
            if 'f_aero' in d_outputs:
                if 'xv' in d_inputs:
                    d_inputs['xv'] += self.qinf * np.asarray(self.adj.computeJacVecLoadsMesh(d_outputs['f_aero'])) # dX = dL_dX^T * dL
                if 'phi' in d_inputs:
                    d_inputs['phi'] += self.qinf * np.asarray(self.adj.computeJacVecLoadsFlow(d_outputs['f_aero'])) # dU = dL_dU^T * dL
        elif mode == 'fwd':
            raise NotImplementedError('DartLoads - forward mode not implemented!\n')

# Aerodynamic coefficients
class DartCoefficients(om.ExplicitComponent):
    """Aerodynamic load coefficients for full aircraft

    Attributes
    ----------
    sol : dart.Newton object
        Direct Newton sovler
    adj : dart.Adjoint object
        Adjoint solver
    """
    def initialize(self):
        self.options.declare('sol', desc='direct solver', recordable=False)
        self.options.declare('adj', desc='adjoint solver', recordable=False)

    def setup(self):
        self.sol = self.options['sol']
        self.adj = self.options['adj']
        # I/O
        self.add_input('aoa', val=0., units='rad', desc='angle of attack', tags=['mphys_input'])
        self.add_input('xv', shape_by_conn=True, desc='aerodynamic volume node coordinates', tags=['mphys_coupling'])
        self.add_input('phi', shape_by_conn=True, desc='flow variables (potential)', tags=['mphys_coupling'])
        self.add_output('cl', val=0., desc='lift coefficient', tags=['mphys_result'])
        self.add_output('cd', val=0., desc='drag coefficient', tags=['mphys_result'])
        # Partials
        self.declare_partials(of=['cl', 'cd'], wrt=['aoa', 'xv', 'phi'])

    def compute(self, inputs, outputs):
        """Get the coefficients for the full aircraft
        """
        # NB: inputs already up-to-date because DartSolver has already been run
        # Update outputs
        outputs['cl'] = self.sol.Cl
        outputs['cd'] = self.sol.Cd

    def compute_partials(self, inputs, partials):
        """Compute the partials gradients of the load coefficients
        """
        # Gradients wrt to angle of attack
        dCfAoa = self.adj.computeGradientCoefficientsAoa()
        partials['cl', 'aoa'] = dCfAoa[0]
        partials['cd', 'aoa'] = dCfAoa[1]
        # Gradients wrt to mesh
        dCfMsh = self.adj.computeGradientCoefficientsMesh()
        partials['cl', 'xv'] = dCfMsh[0]
        partials['cd', 'xv'] = dCfMsh[1]
        # Gradients wrt to flow variables
        dCfPhi = self.adj.computeGradientCoefficientsFlow()
        partials['cl', 'phi'] = dCfPhi[0]
        partials['cd', 'phi'] = dCfPhi[1]

# Aerodynamic writer
class DartWriter(om.ExplicitComponent):
    """Write solution to disk

    Attributes
    ----------
    sol : dart.Newton object
        Direct Newton sovler
    wrt : dart.MshExport object
        Mesh and solution data writer
    sname : string
        Name of scenario
    sall : bool
        Whether to save results at each iteration to different files
    it : int
        Optimization iteration count
    """
    def initialize(self):
        self.options.declare('sol', desc='direct solver', recordable=False)
        self.options.declare('wrt', desc='data writer', recordable=False)
        self.options.declare('snam', desc='name of the scenario')
        self.options.declare('sall', desc='whether to save results at each iteration to different files')

    def setup(self):
        self.sol = self.options['sol']
        self.wrt = self.options['wrt']
        self.snam = self.options['snam'] if self.options['snam'] is not None else 'default'
        self.sall = self.options['sall']
        self.it = 0 # optimization iteration count

    def compute(self, inputs, outputs):
        """Write to disk
        """
        # Create suffix
        suffix = '_{:s}'.format(self.snam)
        if self.sall:
            suffix += '_{:04d}'.format(self.it)
        self.it += 1
        # Write mesh to disk (.msh only)
        if self.wrt.type() == 1:
            self.wrt.save(self.sol.pbl.msh.name + suffix)
        # Write solution to disk
        self.sol.save(self.wrt, suffix)

# Aerodynamic coupling group
class DartGroup(om.Group):
    """Aerodynamic coupling group
    Integrate the aerodynamic computations in an aero-structural coupling procedure

    Components
    ----------
    - Mesh morpher
    - Direct and adjoint solvers
    - Loads computation
    """
    def initialize(self):
        self.options.declare('qinf', desc='freestream dynamic pressure')
        self.options.declare('bnd', desc='moving body', recordable=False)
        self.options.declare('sol', desc='direct solver', recordable=False)
        self.options.declare('mrf', desc='mesh morpher', recordable=False)
        self.options.declare('adj', desc='adjoint solver', recordable=False)
        self.options.declare('rerr', desc='raise AnalysisError when solver not fully converged')

    def setup(self):
        # Components
        if self.options['mrf'] is None:
            self.add_subsystem('morpher', DartDummyMorpher(dim=self.options['sol'].pbl.nDim, msh=self.options['sol'].pbl.msh), promotes_inputs=['x_aero'], promotes_outputs=['xv'])
        else:
            self.add_subsystem('morpher', DartMorpher(dim=self.options['sol'].pbl.nDim, bnd=self.options['bnd'], mrf=self.options['mrf'], adj=self.options['adj']), promotes_inputs=['x_aero'], promotes_outputs=['xv'])
        self.add_subsystem('solver', DartSolver(sol=self.options['sol'], adj=self.options['adj'], rerr=self.options['rerr']), promotes_inputs=['aoa', 'xv'], promotes_outputs=['phi'])
        if self.options['qinf'] is not None:
            self.add_subsystem('loads', DartLoads(qinf=self.options['qinf'], bnd=self.options['bnd'], adj=self.options['adj']), promotes_inputs=['xv', 'phi'], promotes_outputs=['f_aero'])

# Aerodynamic post-coupling group
class DartPostGroup(om.Group):
    """Aerodynamic post-coupling group
    Update the aerodynamic load coefficients and write solution to disk

    Components
    ----------
    - Coefficients
    - Writer
    """
    def initialize(self):
        self.options.declare('sol', desc='direct solver', recordable=False)
        self.options.declare('adj', desc='adjoint solver', recordable=False)
        self.options.declare('wrt', desc='data writer', recordable=False)
        self.options.declare('snam', desc='name of the scenario')
        self.options.declare('sall', desc='whether to save results at each iteration to different files')

    def setup(self):
        # Components
        self.add_subsystem('coeff', DartCoefficients(sol=self.options['sol'], adj=self.options['adj']), promotes_inputs=['aoa', 'xv', 'phi'], promotes_outputs=['cl', 'cd'])
        self.add_subsystem('writer', DartWriter(sol=self.options['sol'], wrt=self.options['wrt'], snam=self.options['snam'], sall=self.options['sall']))

# Builder
class DartBuilder(Builder):
    """Dart builder for MPHYS

    Attributes
    ----------
    __dim : int
        Dimension of the problem
    __qinf : float
        Freestream dynamic pressure
    __wrt : tbox.MshExport object
        Mesh and solution data writer
    __mrf : tbox.MshDeform object
        Mesh morpher
    __sol : dart.Newton object
        Direct Newton solver
    __adj : dart.Adjoint object
        Adjoint solver
    """
    def __init__(self, cfg, scenario='aerodynamic', task='analysis', saveAll=False, raiseError=False):
        """Instantiate the builder and save the parameters and options

        Parameters
        ----------
        cfg : dict
            Dictonnary of parameters to configure DART
        scenario : string, optional
            Type of scenario (available: aerodynamic, aerostructural) (default: aerodynamic)
        task : string, optional
            Type of task (available: analysis, optimization) (default: analysis)
        saveAll : bool, optional
            Whether to write results to different files at each iteration (default: False, results file will be overwritten)
        raiseError : bool, optional
            Whether to raise an openMDAO.AnalysisError when the solver does not fully converge (default: False)
            As of openMDAO V3.9.2, AnalysisError are only handled by pyOptSparse
        """
        self.__cfg = cfg
        self.__scn = scenario
        self.__tsk = task
        self.__sall = saveAll
        self.__rerr = raiseError

    def initialize(self, comm):
        """Instantiate and initialize all the solver components

        Parameters
        ----------
        comm: mpi4py.MPI.Comm object
            MPI communicator
        """
        from .core import initDart
        def _init():
            self.__dim, self.__qinf, _, self.__wrt, self.__mrf, _, self.__bnd, self.__sol, self.__adj = initDart(self.__cfg, scenario=self.__scn, task=self.__tsk)
        # If n MPI processes, init DART on rank=0,...,n-1 sequentially to avoid issues with mesh IO
        # If mpi4py is not available or only 1 MPI process, init DART as usual
        try:
            from mpi4py import MPI
            # If several processes are provided by OpenMDAO, DART is requested to run with MPI, which is not supported
            if comm.size > 1:
                raise RuntimeError('DartBuiler.initialize - DART cannot be run using MPI, except for MultiPoint analysis!\n')
            # Else, we are running a MultiPoint analysis and need to ID the actual process
            true_comm = MPI.COMM_WORLD
            if true_comm.size > 1:
                if true_comm.rank != 0:
                    true_comm.recv(source=true_comm.rank-1)
                _init()
                if true_comm.rank != true_comm.size-1:
                    true_comm.send([], dest=true_comm.rank+1)
                true_comm.barrier()
            else:
                _init()
        except:
            _init()

    def get_mesh_coordinate_subsystem(self, scenario_name=None):
        """Return openMDAO component to get the initial surface mesh coordinates
        """
        return DartMesh(dim=self.__dim, bnd=self.__bnd)

    def get_coupling_group_subsystem(self, scenario_name=None):
        """Return openMDAO group containing the morpher, solver and loads
        """
        return DartGroup(qinf=self.__qinf, bnd=self.__bnd, sol=self.__sol, mrf=self.__mrf, adj=self.__adj, rerr=self.__rerr)

    def get_post_coupling_subsystem(self, scenario_name=None):
        """Return openMDAO group that computes the aero coefficients and writes data to disk
        """
        return DartPostGroup(sol=self.__sol, adj=self.__adj, wrt=self.__wrt, snam=scenario_name, sall=self.__sall)

    def get_number_of_nodes(self):
        """Return the number of surface nodes
        """
        return len(self.__bnd.nodes)
