from matplotlib import pyplot as plt
import numpy as np

class GroupConstructor():
    """ ---------- Group Constructor ----------
        
        Builds single vectors elements out of the 3 groups given in input.
    """
    def __init__(self) -> None:
        pass

    def mergeVectors(self,dataIn, mapMesh, nFactor):
        """Merges 3 groups upper/lower sides and wake.
        """
        
        for k in range(len(dataIn)):

            if len(dataIn[k]) == 2: # Airfoil case.
            
                xx_up, dv_up, vtExt_up, _, alpha_up = self.__getBLcoordinates(dataIn[k][0])
                xx_lw, dv_lw, vtExt_lw, _, alpha_lw = self.__getBLcoordinates(dataIn[k][1])

            else: # Wake case

                xx_wk, dv_wk, vtExt_wk, _, alpha_wk = self.__getBLcoordinates(dataIn[k])

        if mapMesh:
            # Save initial mesh.
            cMesh        = np.concatenate((dataIn[0][0][:,0], dataIn[0][1][1:,0], dataIn[1][:,0]))
            cMeshbNodes  = [0, len(dataIn[0][0][:,0]), len(dataIn[0][0][:,0]) + len(dataIn[0][1][1:,0])]
            cxx          = np.concatenate((xx_up, xx_lw[1:], xx_wk))
            cvtExt       = np.concatenate((vtExt_up, vtExt_lw[1:], vtExt_wk))
            cxxExt       = np.concatenate((dataIn[0][0][:,8], dataIn[0][1][1:,8],dataIn[1][:,8]))

            # Create fine mesh.
            fMeshUp      = self.createFineMesh(dataIn[0][0][:,0], nFactor)
            fMeshLw      = self.createFineMesh(dataIn[0][1][:,0], nFactor)
            fMeshWk      = self.createFineMesh(dataIn[1][:,0]   , nFactor)
            fMesh        = np.concatenate((fMeshUp, fMeshLw[1:], fMeshWk))
            fMeshbNodes  = [0, len(fMeshUp), len(fMeshUp) + len(fMeshLw[1:])]

            fxx_up       = self.interpolateToFineMesh(xx_up, fMeshUp, nFactor)
            fxx_lw       = self.interpolateToFineMesh(xx_lw, fMeshLw, nFactor)
            fxx_wk       = self.interpolateToFineMesh(xx_wk, fMeshWk, nFactor)
            fxx          = np.concatenate((fxx_up, fxx_lw[1:], fxx_wk))

            fvtExt_up    = self.interpolateToFineMesh(vtExt_up, fMeshUp, nFactor)
            fvtExt_lw    = self.interpolateToFineMesh(vtExt_lw, fMeshLw, nFactor)
            fvtExt_wk    = self.interpolateToFineMesh(vtExt_wk, fMeshWk, nFactor)
            fvtExt       = np.concatenate((fvtExt_up, fvtExt_lw[1:], fvtExt_wk))

            fMe_up       = self.interpolateToFineMesh(dataIn[0][0][:,5], fMeshUp, nFactor)
            fMe_lw       = self.interpolateToFineMesh(dataIn[0][1][:,5], fMeshLw, nFactor)
            fMe_wk       = self.interpolateToFineMesh(dataIn[1][:,5]   , fMeshWk, nFactor)
            fMe          = np.concatenate((fMe_up, fMe_lw[1:], fMe_wk))

            frho_up      = self.interpolateToFineMesh(dataIn[0][0][:,6], fMeshUp, nFactor)
            frho_lw      = self.interpolateToFineMesh(dataIn[0][1][:,6], fMeshLw, nFactor)
            frho_wk      = self.interpolateToFineMesh(dataIn[1][:,6]   , fMeshWk, nFactor)
            frho         = np.concatenate((frho_up, frho_lw[1:], frho_wk))

            fdStarExt_up = self.interpolateToFineMesh(dataIn[0][0][:,7], fMeshUp, nFactor)
            fdStarExt_lw = self.interpolateToFineMesh(dataIn[0][1][:,7], fMeshLw, nFactor)
            fdStarExt_wk = self.interpolateToFineMesh(dataIn[1][:,7]   , fMeshWk, nFactor)
            
            fxxExt_up    = self.interpolateToFineMesh(dataIn[0][0][:,8], fMeshUp, nFactor)
            fxxExt_lw    = self.interpolateToFineMesh(dataIn[0][1][:,8], fMeshLw, nFactor)
            fxxExt_wk    = self.interpolateToFineMesh(dataIn[1][:,8]   , fMeshWk, nFactor)
            
            fdStarext    = np.concatenate((fdStarExt_up, fdStarExt_lw[1:], fdStarExt_wk))
            fxxext       = np.concatenate((fxxExt_up, fxxExt_lw[1:], fxxExt_wk))

            fdv          = np.concatenate((dv_up, dv_lw[1:], dv_wk))

            fMeshDict    = {'globCoord': fMesh, 'bNodes': fMeshbNodes, 'locCoord': fxx,
                            'vtExt': fvtExt, 'Me': fMe, 'rho': frho, 'deltaStarExt': fdStarext, 'xxExt': fxxext, 'dv': fdv}

            cMe          = np.concatenate((dataIn[0][0][:,5], dataIn[0][1][1:,5], dataIn[1][:,5]))
            cRho         = np.concatenate((dataIn[0][0][:,6], dataIn[0][1][1:,6], dataIn[1][:,6]))
            cdStarExt    = np.concatenate((dataIn[0][0][:,7], dataIn[0][1][1:,7], dataIn[1][:,7]))
            cMeshDict    = {'globCoord': cMesh, 'bNodes': cMeshbNodes, 'locCoord': cxx,
                            'vtExt': cvtExt, 'Me': cMe, 'rho': cRho, 'deltaStarExt': cdStarExt, 'xxExt': cxxExt, 'dv': fdv}

            dataUpper = np.zeros((len(fMeshUp), 0))
            dataLower = np.zeros((len(fMeshLw), 0))
            dataWake  = np.zeros((len(fMeshWk), 0))
            for iParam in range(len(dataIn[0][0][0,:])):
                interpParamUp = self.interpolateToFineMesh(dataIn[0][0][:,iParam], fMeshUp, nFactor)
                dataUpper     = np.column_stack((dataUpper, interpParamUp))
                interpParamLw = self.interpolateToFineMesh(dataIn[0][1][:,iParam], fMeshLw, nFactor)
                dataLower     = np.column_stack((dataLower, interpParamLw))
                interpParamWk = self.interpolateToFineMesh(dataIn[1][:,iParam]   , fMeshWk, nFactor)
                dataWake      = np.column_stack((dataWake, interpParamWk))
            
            dataLower = np.delete(dataLower, 0, axis = 0) # Remove stagnation point doublon. 

        else:

            Mesh         = np.concatenate((dataIn[0][0][:,0], dataIn[0][1][:,0], dataIn[1][:,0]))
            Meshy         = np.concatenate((dataIn[0][0][:,1], dataIn[0][1][:,1], dataIn[1][:,1]))
            Meshz         = np.concatenate((dataIn[0][0][:,2], dataIn[0][1][:,2], dataIn[1][:,2]))
            MeshbNodes   = [0, len(dataIn[0][0][:,0]), len(dataIn[0][0][:,0]) + len(dataIn[0][1][:,0]), len(dataIn[0][0][:,0]) + len(dataIn[0][1][:,0]) + len(dataIn[1][:,0])]

            xx           = np.concatenate((xx_up, xx_lw, xx_wk))
            vtExt        = np.concatenate((vtExt_up, vtExt_lw, vtExt_wk))

            vx = np.concatenate((dataIn[0][0][:,3], dataIn[0][1][:,3], dataIn[1][:,3]))
            vy = np.concatenate((dataIn[0][0][:,4], dataIn[0][1][:,4], dataIn[1][:,4]))
            vz = np.concatenate((dataIn[0][0][:,5], dataIn[0][1][:,5], dataIn[1][:,5]))

            alpha        = np.concatenate((alpha_up, alpha_lw, alpha_wk))
            dv           = np.concatenate((dv_up, dv_lw, dv_wk))

            Me           = np.concatenate((dataIn[0][0][:,5], dataIn[0][1][:,5], dataIn[1][:,5]))
            rho          = np.concatenate((dataIn[0][0][:,6], dataIn[0][1][:,6], dataIn[1][:,6]))
            dStarext     = np.concatenate((dataIn[0][0][:,7], dataIn[0][1][:,7], dataIn[1][:,7]))
            xxExt        = np.concatenate((dataIn[0][0][:,8], dataIn[0][1][:,8], dataIn[1][:,8]))

            fMeshDict    = {'globCoord': Mesh, 'yGlobCoord': Meshy, 'zGlobCoord': Meshz, 'bNodes': MeshbNodes, 'locCoord': xx,
                            'vtExt': vtExt, 'vx': vx, 'vy': vy, 'vz': vz, 'Me': Me, 'rho': rho, 'deltaStarExt': dStarext, 'xxExt': xxExt, 'dv': dv}

            cMeshDict    = fMeshDict.copy()

            dataUpper    = dataIn[0][0]
            dataLower    = dataIn[0][1][1:, :]
            dataWake     = dataIn[1]
        
        data = np.concatenate((dataUpper, dataLower, dataWake), axis = 0)

        return fMeshDict, cMeshDict, data

    def __getBLcoordinates(self,data):
        """Transform the reference coordinates into airfoil coordinates
        """
        nN    = len(data[:,0])
        nE    = nN-1            # Nbr of element along the surface.
        vt    = np.zeros(nN)    # Outer flow velocity. 
        xx    = np.zeros(nN)    # Position along the surface of the airfoil.
        dx    = np.zeros(nE)    # Distance along two nodes.
        dv    = np.zeros(nE)    # Speed difference according to element.
        alpha = np.zeros(nE)    # Angle of the element for the rotation matrix.

        for i in range(0, nE):

            alpha[i] = np.arctan2((data[i+1, 1] - data[i, 1]), (data[i+1, 0]-data[i, 0]))
            dx[i]    = np.sqrt((data[i+1, 0] - data[i,0])**2 + (data[i+1, 1]-data[i, 1])**2)
            xx[i+1]  = dx[i] + xx[i]
            vt[i]    = (data[i, 3] * np.cos(alpha[i]) + data[i ,4] * np.sin(alpha[i]))          # Tangential speed at node 1 element i
            vt[i+1]  = (data[i+1, 3] * np.cos(alpha[i]) + data[i+1, 4] * np.sin(alpha[i]))      # Tangential speed at node 2 element i
            dv[i]    = (vt[i+1] - vt[i]) / dx[i]                                                # Velocity gradient according to element i.
                                                                                                #             central scheme with half point
        
        return xx, dv, vt, nN, alpha

    def interpolateToFineMesh(self, cVector, fMesh, nFactor):
        
        fVector = np.zeros(len(fMesh)-1)
        for cPoint in range(len(cVector) - 1):
            dv = cVector[cPoint + 1] - cVector[cPoint]

            for fPoint in range(nFactor):
                fVector[nFactor * cPoint + fPoint] = cVector[cPoint] + fPoint * dv / nFactor
        
        fVector = np.append(fVector, cVector[-1])

        return fVector

    def createFineMesh(self, cMesh, nFactor):

        fMesh = []
        for iPoint in range(len(cMesh) - 1):

            dx = cMesh[iPoint + 1] - cMesh[iPoint]

            for iRefinedPoint in range(nFactor):

                fMesh = np.append(fMesh, cMesh[iPoint] + iRefinedPoint * dx / nFactor)

        fMesh = np.append(fMesh, cMesh[-1])
        return fMesh
