#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## VII Coupler
# Paul Dechamps

import fwk
from fwk.coloring import ccolors

from math import pi as M_PI
import numpy as np

class Data:
    def __init__(self, msh,  visc=0, PosSections= None):

        self.Points, self.Cg = self.GetMesh(msh)

        if visc:
            self.Xp, self.Yp, self.Zp, self.sortedRows, self.Xc, self.Yc, self.Zc, self.sortedElems = self.SortMesh(self.Points, self.Cg)

            # Creating sections
            self.Sections = self.CreateSections(PosSections)

            print(self.Sections)


    def GetMesh(self, msh):

        data = [np.zeros((0,4)) for _ in range(2)]
        cg = [np.zeros((0,4)) for _ in range(2)]

        alreadysaid = ['']
        for e in msh.elems:
            if e.ptag.name not in alreadysaid:
                alreadysaid.append(e.ptag.name)

        for e in msh.elems:
            # Compute cg position
            cg_pt = np.zeros(3)
            for n in e.nodes:
                cg_pt += [n.pos[0], n.pos[1], n.pos[2]]
            cg_pt/=e.nodes.size()
            cg_pt = np.hstack((cg_pt, e.no))

            if e.ptag.name == "wake":
                for nw in e.nodes:
                    if nw.row not in data[1][:,3]:
                        data[1] = np.vstack((data[1], [nw.pos[0], nw.pos[1], nw.pos[2], nw.row]))
                cg[1] = np.vstack((cg[1], cg_pt))
            elif e.ptag.name == 'wing':
                for nw in e.nodes:
                    if nw.row not in data[0][:,3]:
                        data[0] = np.vstack((data[0], [nw.pos[0], nw.pos[1], nw.pos[2], nw.row]))
                cg[0] = np.vstack((cg[0], cg_pt))
            elif e.ptag.name == 'wing_':
                for nw in e.nodes:
                    if nw.row not in data[0][:,3]:
                        data[0] = np.vstack((data[0], [-nw.pos[0], nw.pos[1], nw.pos[2], nw.row]))
                cg_pt[0] = -cg_pt[0]
                cg[0] = np.vstack((cg[0], cg_pt))
        
        return data, cg

    def SortMesh(self, data, cg):

        data = data.copy()

        data[0] = data[0][data[0][:,1]<=1, :]

        cg = data.copy()

        cg[0] = cg[0][cg[0][:,1]<=1, :]

        sections = data[0][:,1]
        for i in range(len(sections)):
            sections[i] = round(sections[i], 3)
        sections = np.unique(sections)

        nChord = [len(data[i][abs(data[i][:,1]-sections[0])<1e-3, 0]) for i in range(len(data))]

        sortedRows = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Xp = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Yp = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Zp = [np.zeros((nChord[i], 0)) for i in range(len(data))]

        Xc = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Yc = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Zc = [np.zeros((nChord[i], 0)) for i in range(len(data))]

        for iReg, reg in enumerate(data):

            for iy, y in enumerate(sections):

                # Find nearest value in the mesh

                idx = (np.abs(reg[:,1]-y).argmin())
                sections[iy] = reg[idx,1]

                sort = reg[abs(reg[:,1]-sections[iy])<1e-3, :]
                sort = sort[sort[:,0].argsort()]

                sortedRows[iReg] = np.column_stack((sortedRows[iReg], sort[:,3]))
                Xp[iReg] = np.column_stack((Xp[iReg], sort[:,0]))
                Yp[iReg] = np.column_stack((Yp[iReg], sort[:,1]))
                Zp[iReg] = np.column_stack((Zp[iReg], sort[:,2]))

        sections_cg = cg[0][:,1]
        for i in range(len(sections_cg)):
            sections_cg[i] = round(sections_cg[i], 3)
        sections_cg = np.unique(sections)

        nChord_cg = [len(cg[i][abs(cg[i][:,1]-sections[0])<1e-3, 0]) for i in range(len(cg))]

        sortedElems = [np.zeros((nChord_cg[i], 0)) for i in range(len(cg))]
        
        for iReg, reg in enumerate(cg):

            for iy, y in enumerate(sections_cg):

                # Find nearest value in the mesh

                idx = (np.abs(reg[:,1]-y).argmin())
                sections_cg[iy] = reg[idx,1]

                sort = reg[abs(reg[:,1]-sections_cg[iy])<1e-3, :]
                sort = sort[sort[:,0].argsort()]
                
                sortedElems[iReg] = np.column_stack((sortedElems[iReg], sort[:,3]))
                Xc[iReg] = np.column_stack((Xc[iReg], sort[:,0]))
                Yc[iReg] = np.column_stack((Yc[iReg], sort[:,1]))
                Zc[iReg] = np.column_stack((Zc[iReg], sort[:,2]))

        return Xp, Yp, Zp, sortedRows, Xc, Yc, Zc, sortedElems

    def CreateSections(self, PosSections):
        Sections = [[np.zeros((len(self.Xp[iReg]), 4)) for iReg in range(2)] for _ in range(len(PosSections))]

        for iSec, y in enumerate(PosSections):
            for iReg in range(2):
                print(self.Yp[iReg][0,:])
                print(np.where(self.Yp[iReg][0,:]==y)[0])
                Sections[iSec][iReg][:,0] = self.Xp[iReg][:,np.where(self.Yp[iReg][0,:]==y)[0][0]]
                Sections[iSec][iReg][:,1] = self.Yp[iReg][:,np.where(self.Yp[iReg][0,:]==y)[0][0]]
                Sections[iSec][iReg][:,2] = self.Zp[iReg][:,np.where(self.Yp[iReg][0,:]==y)[0][0]]
                Sections[iSec][iReg][:,3] = self.sortedRows[iReg][:,np.where(self.Yp[iReg][0,:]==y)[0][0]]

        print(Sections[0][0])
        return Sections


        