/* ----------------------------------------------------------------------------------- */
/*                                                                                     */
/*                       Coupled integral boundary layer solver                        */
/*                                                                                     */
/*  Université de Liège                                                                */
/*  February 2022                                                                      */
/*  Author: Paul Dechamps                                                              */
/*  v0.1                                                                               */
/* ----------------------------------------------------------------------------------- */

/* --- Project includes -------------------------------------------------------------- */

#include "wViscSolver.h"
#include "wBLRegion.h"
#include "wTimeSolver.h"
#include <iomanip>

#define ANSI_COLOR_RED "\x1b[1;31m"
#define ANSI_COLOR_GREEN "\x1b[1;32m"
#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

/* --- Namespaces -------------------------------------------------------------------- */

using namespace vii;

/* ----------------------------------------------------------------------------------- */

/**
 * @brief Viscous solver and boundary layer(s) initialization
 */
ViscSolver::ViscSolver(double _Re, double _Minf, double _CFL0, unsigned int nSections, double _Span, unsigned int _verbose)
{

    PrintBanner();

    // Freestream parameters.
    Re = _Re;
    CFL0 = _CFL0;

    verbose = _verbose;

    // Initialzing boundary layer
    Sections.resize(nSections);
    for (size_t iSec=0; iSec < nSections; ++iSec)
    {
        for (size_t i = 0; i < 3; ++i)
        {
            BLRegion *region = new BLRegion();
            Sections[iSec].push_back(region);
        }
        Sections[iSec][0]->name = "upper";
        Sections[iSec][1]->name = "lower";
        Sections[iSec][2]->name = "wake";
    }

    Cdt_sec.resize(Sections.size(), 0.);
    Cdf_sec.resize(Sections.size(), 0.);
    Cdp_sec.resize(Sections.size(), 0.);
    Cdt = 0.; Cdf = 0.; Cdp = 0.;
    Span = _Span;

    // Pre/post processing flags
    solverExitCode = 0;
    stagPtMvmt.resize(Sections.size(), false);

    // Temporal solver initialization.
    tSolver = new TimeSolver(_CFL0, _Minf, Re);

    // User input numerical parameters.
    std::cout << "--- Viscous solver setup ---\n"
              << "Reynolds number: " << Re << " \n"
              << "Freestream Mach number: " << _Minf << " \n"
              << "Initial CFL number: " << CFL0 << " \n"
              << std::endl;
    std::cout << "Boundary layer initialized" << std::endl;
}

/**
 * @brief Viscous solver and subsequent dependancies destruction
 */
ViscSolver::~ViscSolver()
{   
    for (size_t iSec=0; iSec<Sections.size(); ++iSec)
    {
        for (size_t i = 0; i < Sections[iSec].size(); ++i)
        {
            delete Sections[iSec][i];
        }
    }
    delete tSolver;
    std::cout << "~ViscSolver()\n";
} // end ~ViscSolver

/**
 * @brief Runs viscous operations. Temporal solver parametrisation is first adapted,
 * Solutions are initialized than time marched towards steady state successively for each points.
 */
int ViscSolver::Run(unsigned int const couplIter)
{   
    bool lockTrans=false;   /* Flagging activation of transition routines */
    int pointExitCode = 0;  /* Output of pseudo time integration (0: converged; -1: unsuccessful, failed; >0: unsuccessful nb iter) */
    solverExitCode = 0;

    double maxMach = 0.;

    for (size_t iSec=0; iSec<Sections.size(); ++iSec)
    {
        tms["0-CheckIn"].start();

        if (verbose>1){std::cout << "Sec " << iSec << " ";}

        // Check for stagnation point movement.
        if (couplIter != 0 && (Sections[iSec][0]->mesh->GetDiffnElm()))
        {
            stagPtMvmt[iSec] = true;
            if (verbose>2) std::cout << "Stagnation point moved by " << Sections[iSec][0]->mesh->GetDiffnElm() << " elements." << std::endl;
        }
        else
            stagPtMvmt[iSec] = false;

        // Set boundary conditions.
        Sections[iSec][0]->xtr = 1.; /* Upper side initial xtr */
        Sections[iSec][1]->xtr = 1.; /* Lower side initial xtr */
        Sections[iSec][2]->xtr = 0.; /* Wake initial xtr (fully turbulent) */

        Sections[iSec][0]->SetStagBC(Re);
        Sections[iSec][1]->SetStagBC(Re);

        tms["0-CheckIn"].stop();

        // Loop over the regions (upper, lower and wake).
        for (size_t iRegion = 0; iRegion < Sections[iSec].size(); ++iRegion)
        {
            if (Sections[iSec][iRegion]->invBnd->GetMaxMach() > maxMach)
                maxMach = Sections[iSec][iRegion]->invBnd->GetMaxMach();

            if (verbose>1) std::cout << Sections[iSec][iRegion]->name << " ";

            // Check if safe mode is required.
            if (couplIter == 0 || Sections[iSec][iRegion]->mesh->GetDiffnElm() != 0 || stagPtMvmt[iSec] == true || solverExitCode != 0)
            {
                tSolver->SetCFL0(1);
                tSolver->SetitFrozenJac(1);
                tSolver->SetinitSln(true);

                if (Sections[iSec][iRegion]->mesh->GetDiffnElm())
                {
                    std::vector<double> xxExtCurr(Sections[iSec][iRegion]->mesh->GetnMarkers(), 0.);
                    for (size_t iPoint=0; iPoint<xxExtCurr.size(); ++iPoint)
                        xxExtCurr[iPoint] = Sections[iSec][iRegion]->mesh->GetLoc(iPoint);
                    Sections[iSec][iRegion]->mesh->SetExt(xxExtCurr);

                    std::vector<double> DeltaStarZeros(Sections[iSec][iRegion]->mesh->GetnMarkers(), 0.0);
                    Sections[iSec][iRegion]->invBnd->SetDeltaStar(DeltaStarZeros);

                    std::vector<double> ueOld(Sections[iSec][iRegion]->mesh->GetnMarkers(), 0.0);
                    for (size_t iPoint=0; iPoint<ueOld.size(); ++iPoint)
                        ueOld[iPoint] = Sections[iSec][iRegion]->invBnd->GetVt(iPoint);
                    Sections[iSec][iRegion]->invBnd->SetVtOld(ueOld);
                }

                // Keyword annoncing iteration safety level.
                if (verbose>1) std::cout << "restart. ";

            }
            else
            {
                tSolver->SetCFL0(CFL0);
                tSolver->SetitFrozenJac(5);
                tSolver->SetinitSln(false);

                // Keyword annoncing iteration safety level.
                if (verbose>1) std::cout << "update. ";
            }

            // Save number of points in each regions.
            Sections[iSec][iRegion]->mesh->prevnMarkers = Sections[iSec][iRegion]->mesh->GetnMarkers();

            // Reset regime.
            lockTrans = false;
            if (iRegion < 2) // Airfoil
                for (size_t k = 0; k < Sections[iSec][iRegion]->mesh->GetnMarkers(); k++)
                    Sections[iSec][iRegion]->Regime[k] = 0;
                    
            else // Wake
            {
                for (size_t k = 0; k < Sections[iSec][iRegion]->mesh->GetnMarkers(); k++)
                    Sections[iSec][iRegion]->Regime[k] = 1;

                // Impose wake boundary condition.
                std::vector<double> UpperCond(Sections[iSec][iRegion]->GetnVar(), 0.);
                std::vector<double> LowerCond(Sections[iSec][iRegion]->GetnVar(), 0.);
                for (size_t k=0; k<UpperCond.size(); ++k)
                {
                    UpperCond[k] = Sections[iSec][0]->U[(Sections[iSec][0]->mesh->GetnMarkers() - 1) * Sections[iSec][0]->GetnVar() + k];
                    LowerCond[k] = Sections[iSec][0]->U[(Sections[iSec][1]->mesh->GetnMarkers() - 1) * Sections[iSec][1]->GetnVar() + k];
                }
                Sections[iSec][iRegion]->SetWakeBC(Re, UpperCond, LowerCond);
                lockTrans = true;
                if (verbose>2) std::cout << "Wake BC imposed." <<std::endl;
            }

            // Loop over points
            for (size_t iPoint = 1; iPoint < Sections[iSec][iRegion]->mesh->GetnMarkers(); ++iPoint)
            {

                // Initialize solution at point
                tSolver->InitialCondition(iPoint, Sections[iSec][iRegion]);
                // Solve equations 
                tms["1-TimeSolver"].start();
                pointExitCode = tSolver->Integration(iPoint, Sections[iSec][iRegion]);
                tms["1-TimeSolver"].stop();

                // Unsucessfull convergence output
                if (pointExitCode < 0)
                {
                    if (verbose>1){std::cout << "Point " << iPoint << ": Convergence terminated with exitcode " << pointExitCode << std::endl;}
                    solverExitCode = -1; // Convergence failed
                }
                else if (pointExitCode > 0 and solverExitCode >=0)
                    solverExitCode = pointExitCode; // Not fully converged
                else
                    if (verbose>2) std::cout << "Point " << iPoint << ": Successful convergence.\n";

                // Transition
                tms["2-Transition"].start();
                if (!lockTrans)
                {
                    // Check if perturbation waves are growing or not.
                    CheckWaves(iPoint, Sections[iSec][iRegion]);

                    // Amplification factor is compared to critical amplification factor.
                    if (Sections[iSec][iRegion]->U[iPoint * Sections[iSec][iRegion]->GetnVar() + 2] >= Sections[iSec][iRegion]->GetnCrit())
                    {
                        AverageTransition(iPoint, Sections[iSec][iRegion], 0);
                        if (verbose>1)
                        {   
                            std::cout << std::fixed;
                            std::cout << std::setprecision(2);
                            std::cout << Sections[iSec][iRegion]->xtr
                                      << " (" << Sections[iSec][iRegion]->transMarker << ") ";
                        }
                        if (verbose>2)
                            std::cout << "" << std::endl;
                        lockTrans = true;
                    }
                }
                tms["2-Transition"].stop();

            } // End for iPoints

            tms["3-Blowing"].start();
            Sections[iSec][iRegion]->ComputeBlwVel();
            tms["3-Blowing"].stop();
            if (verbose>2){"Blowing velocities of " + Sections[iSec][iRegion]->name + " side computed.";}

        } // End for iRegion
        if (verbose>1) std::cout << "\n";
    } // End for iSections

    for (size_t i=0; i<Sections.size(); ++i)
    {
        ComputeSectionalDrag(Sections[i], i);
    }
    ComputeTotalDrag();

    // Compute mean transition locations
    std::vector<double> meanTransitions(2, 0.);
    for (size_t iSec=0; iSec<Sections.size(); ++iSec)
        for (size_t iReg=0; iReg<2; ++iReg)
            meanTransitions[iReg] += Sections[iSec][iReg]->xtr;
    for (size_t iReg=0; iReg<meanTransitions.size(); ++iReg)
        meanTransitions[iReg]/=Sections.size();

    if (verbose > 0)
    {
        if (verbose > 1) std::cout << "\n" << std::endl;
        std::cout << std::fixed << std::setprecision(2)
                  << "Viscous solver for Re " << Re/1e6 << "e6, Mach max "
                  << maxMach
                  << std::endl;

        std::cout << std::setw(8) << "xTr_Up"
                  << std::setw(12) << "xTr_Lw"
                  << std::setw(12) << "Cd" << std::endl;
        std::cout << "-----------------------------------\n";
        std::cout << std::fixed << std::setprecision(5);
        std::cout << std::setw(8) << meanTransitions[0]
                  << std::setw(12) << meanTransitions[1]
                  << std::setw(12) << Cdt << "\n";

        if (verbose > 2)
        {
            std::cout << "\n" << std::endl;
            std::cout << "Viscous solver CPU" << std::endl
                        << tms;
        }
        std::cout << "\n" << std::endl;
    }
    return solverExitCode; // exit code
}

/**
 * @brief Check whether or not instabilities are growing in the laminar boundary layer.
 */
void ViscSolver::CheckWaves(size_t iPoint, BLRegion *bl)
{

    double logRet_crit = 2.492 * std::pow(1 / (bl->Hk[iPoint] - 1), 0.43) + 0.7 * (std::tanh(14 * (1 / (bl->Hk[iPoint] - 1)) - 9.24) + 1.0);

    if (bl->Ret[iPoint] > 0)
    { // Avoid log of negative number.
        if (std::log10(bl->Ret[iPoint]) <= logRet_crit - 0.08)
            bl->U[iPoint * bl->GetnVar() + 2] = 0; // Set N to 0 at that point if waves do not grow.
    }
}

/**
 * @brief Computes turbulent solution @ the transition point and averages solutions.
 */
void ViscSolver::AverageTransition(size_t iPoint, BLRegion *bl, int forced)
{
    // Averages solution on transition marker.

    unsigned int nVar = bl->GetnVar();

    // Save laminar solution.

    std::vector<double> lamSol(nVar, 0.0);
    for (size_t k = 0; k < lamSol.size(); ++k)
    {
        lamSol[k] = bl->U[iPoint * nVar + k];
    }

    // Set up turbulent portion boundary condition.

    bl->transMarker = iPoint; // Save transition marker.

    // Loop over remaining points in the region.
    for (size_t k = iPoint; k < bl->mesh->GetnMarkers(); ++k)
        bl->Regime[k] = 1; // Set Turbulent regime.

    // Compute transition location.
    bl->xtr = (bl->GetnCrit() - bl->U[(iPoint - 1) * nVar + 2]) * ((bl->mesh->Getx(iPoint) - bl->mesh->Getx(iPoint - 1)) / (bl->U[iPoint * nVar + 2] - bl->U[(iPoint - 1) * nVar + 2])) + bl->mesh->Getx(iPoint - 1);

    // Percentage of the subsection corresponding to a turbulent flow.
    double avTurb = (bl->mesh->Getx(iPoint) - bl->xtr) / (bl->mesh->Getx(iPoint) - bl->mesh->Getx(iPoint - 1));
    double avLam = 1. - avTurb;

    // Impose boundary condition.
    double Cteq_trans;
    bl->closSolver->TurbulentClosures(Cteq_trans, bl->U[(iPoint - 1) * nVar + 0], bl->U[(iPoint - 1) * nVar + 1], 0.03, bl->Ret[iPoint - 1], bl->invBnd->GetMe(iPoint - 1), bl->name);
    bl->Cteq[iPoint - 1] = Cteq_trans;
    bl->U[(iPoint - 1) * nVar + 4] = 0.7 * bl->Cteq[iPoint - 1];

    // Avoid starting with ill conditioned IC for turbulent computation. (Regime was switched above).
    // These initial guess values do not influence the converged solution.
    bl->U[iPoint * nVar + 4] = bl->U[(iPoint - 1) * nVar + 4]; // IC of transition point = turbulent BC (imposed @ previous point).
    bl->U[iPoint * nVar + 1] = 1.515;                          // Because we expect a significant drop of the shape factor.

    /* Solve point in turbulent configuration */
    int exitCode = tSolver->Integration(iPoint, bl);

    if (exitCode != 0 && verbose>1)
        std::cout << "Warning: Transition point turbulent computation terminated with exit code " << exitCode << std::endl;
    else if(verbose>2)
        std::cout << "Sucessfull transition point turbulent computation.";

    // Average both solutions (Now solution @ iPoint is the turbulent solution).
    bl->U[iPoint * nVar + 0] = avLam * lamSol[0] + avTurb * bl->U[(iPoint)*nVar + 0];   /* Theta */
    bl->U[iPoint * nVar + 1] = avLam * lamSol[1] + avTurb * bl->U[(iPoint)*nVar + 1];   /* H */
    bl->U[iPoint * nVar + 2] = 9.;                                                      /* N */
    bl->U[iPoint * nVar + 3] = avLam * lamSol[3] + avTurb * bl->U[(iPoint)*nVar + 3];   /* ue */
    bl->U[iPoint * nVar + 4] = avTurb * bl->U[(iPoint)*nVar + 4];                       /* Ct */

    // Recompute closures. (The turbulent BC @ iPoint - 1 does not influence laminar closure relations).
    std::vector<double> lamParam(8, 0.);
    bl->closSolver->LaminarClosures(lamParam, bl->U[(iPoint - 1) * nVar + 0], bl->U[(iPoint - 1) * nVar + 1], bl->Ret[iPoint - 1], bl->invBnd->GetMe(iPoint - 1), bl->name);
    bl->Hstar[iPoint - 1] = lamParam[0];
    bl->Hstar2[iPoint - 1] = lamParam[1];
    bl->Hk[iPoint - 1] = lamParam[2];
    bl->Cf[iPoint - 1] = lamParam[3];
    bl->Cd[iPoint - 1] = lamParam[4];
    bl->Delta[iPoint - 1] = lamParam[5];
    bl->Cteq[iPoint - 1] = lamParam[6];
    bl->Us[iPoint - 1] = lamParam[7];

    std::vector<double> turbParam(8, 0.);
    bl->closSolver->TurbulentClosures(turbParam, bl->U[(iPoint)*nVar + 0], bl->U[(iPoint)*nVar + 1], bl->U[(iPoint)*nVar + 4], bl->Ret[iPoint], bl->invBnd->GetMe(iPoint), bl->name);
    bl->Hstar[iPoint] = turbParam[0];
    bl->Hstar2[iPoint] = turbParam[1];
    bl->Hk[iPoint] = turbParam[2];
    bl->Cf[iPoint] = turbParam[3];
    bl->Cd[iPoint] = turbParam[4];
    bl->Delta[iPoint] = turbParam[5];
    bl->Cteq[iPoint] = turbParam[6];
    bl->Us[iPoint] = turbParam[7];
}

/**
 * @brief Friction, pressure and total drag computation of the configuration.
 * 
 * Cdtot = theta * Ue^((H+5)/2).
 * Cdf = integral(Cf dx)_upper + integral(Cf dx)_lower.
 * Cdp = Cdtot - Cdf.
 */
void ViscSolver::ComputeSectionalDrag(std::vector<BLRegion *> const &bl, int i)
{

    unsigned int nVar = bl[0]->GetnVar();
    unsigned int lastWkPt = (bl[2]->mesh->GetnMarkers() - 1) * nVar;

    // Normalize friction and dissipation coefficients.
    std::vector<double> frictioncoeff_upper(bl[0]->mesh->GetnMarkers(), 0.0);
    for (size_t k = 0; k < frictioncoeff_upper.size(); ++k)
        frictioncoeff_upper[k] = bl[0]->invBnd->GetVt(k) * bl[0]->invBnd->GetVt(k) * bl[0]->Cf[k];

    std::vector<double> frictioncoeff_lower(bl[1]->mesh->GetnMarkers(), 0.0);
    for (size_t k = 0; k < frictioncoeff_lower.size(); ++k)
        frictioncoeff_lower[k] = bl[1]->invBnd->GetVt(k) * bl[1]->invBnd->GetVt(k) * bl[1]->Cf[k];

    // Compute friction drag coefficient.
    double Cdf_upper = 0.0;
    for (size_t k = 1; k < bl[0]->mesh->GetnMarkers(); ++k)
        Cdf_upper += (frictioncoeff_upper[k] + frictioncoeff_upper[k - 1]) * (bl[0]->mesh->Getx(k) - bl[0]->mesh->Getx(k - 1)) / 2;
    double Cdf_lower = 0.0;
    for (size_t k = 1; k < bl[1]->mesh->GetnMarkers(); ++k)
        Cdf_lower += (frictioncoeff_lower[k] + frictioncoeff_lower[k - 1]) * (bl[1]->mesh->Getx(k) - bl[1]->mesh->Getx(k - 1)) / 2;

    Cdf_sec[i] = Cdf_upper + Cdf_lower; // No friction in the wake

    // Compute total drag coefficient.
    Cdt_sec[i] = 2 * bl[2]->U[lastWkPt + 0] * pow(bl[2]->U[lastWkPt + 3], (bl[2]->U[lastWkPt + 1] + 5) / 2);

    // Compute pressure drag coefficient.
    Cdp_sec[i] = Cdt_sec[i] - Cdf_sec[i];
}

void ViscSolver::ComputeTotalDrag()
{

    Cdt = 0.; Cdf = 0.; Cdp = 0.;

    if (Sections.size() == 1 || Span == 0.)
    {
        Cdt = Cdt_sec[0];
        Cdf = Cdf_sec[0];
        Cdp = Cdp_sec[0];
    }

    else
    {
        Cdt += (Sections[0][0]->mesh->Getz(0) - 0) * Cdt_sec[0];
        Cdp += (Sections[0][0]->mesh->Getz(0) - 0) * Cdp_sec[0];
        Cdf += (Sections[0][0]->mesh->Getz(0) - 0) * Cdf_sec[0];
        double dz = 0.;
        for(size_t k=1; k<Sections.size(); ++k)
        {
            dz = (Sections[k][0]->mesh->Getz(0) - Sections[k-1][0]->mesh->Getz(0));
            Cdt += dz * Cdt_sec[k];
            Cdp += dz * Cdp_sec[k];
            Cdf += dz * Cdf_sec[k];
        }
        Cdt /= Span; Cdp /= Span; Cdf /= Span;
    }
    
}

void ViscSolver::SetGlobMesh(size_t iSec, size_t iRegion, std::vector<double> _x, std::vector<double> _y, std::vector<double> _z)
{
    if (verbose>2){std::cout << "Mesh of " << _x.size() << " elements on " << Sections[iSec][iRegion]->name << " side. ";}
    Sections[iSec][iRegion]->SetMesh(_x, _y, _z);
    if (verbose>3){std::cout << "Mesh mapped to local frame of reference." << std::endl;}
}

void ViscSolver::SetVelocities(size_t iSec, size_t iRegion, std::vector<double> vx, std::vector<double> vy, std::vector<double> vz)
{
    if (vx.size() != vy.size() || vy.size() != vz.size() || vx.size() != vz.size())
    {
        throw std::runtime_error("dart::ViscSolver Wrong velocity vector sizes.");
    }
    if (vx.size() != Sections[iSec][iRegion]->mesh->GetnMarkers())
    {
        throw std::runtime_error("dart::ViscSolver Velocity vector is not coherent with the mesh.");
    }

    std::vector<double> Vt(vx.size(), 0.0);

    for (size_t iPoint=0; iPoint<vx.size()-1; ++iPoint)
    {
        Vt[iPoint] = vx[iPoint]*std::cos(Sections[iSec][iRegion]->mesh->GetAlpha(iPoint)) + vy[iPoint]*std::sin(Sections[iSec][iRegion]->mesh->GetAlpha(iPoint));
        Vt[iPoint+1] = vx[iPoint+1]*std::cos(Sections[iSec][iRegion]->mesh->GetAlpha(iPoint)) + vy[iPoint+1]*std::sin(Sections[iSec][iRegion]->mesh->GetAlpha(iPoint));
    }

    Sections[iSec][iRegion]->invBnd->SetVt(Vt);
}

void ViscSolver::SetMe(size_t iSec, size_t iRegion, std::vector<double> _Me)
{
    if (_Me.size() != Sections[iSec][iRegion]->mesh->GetnMarkers())
        throw std::runtime_error("dart::ViscSolver Mach number vector is not coherent with the mesh.");
    Sections[iSec][iRegion]->invBnd->SetMe(_Me);
}

void ViscSolver::SetRhoe(size_t iSec, size_t iRegion, std::vector<double> _Rhoe)
{
    if (_Rhoe.size() != Sections[iSec][iRegion]->mesh->GetnMarkers())
        throw std::runtime_error("dart::ViscSolver Density vector is not coherent with the mesh.");
    Sections[iSec][iRegion]->invBnd->SetRhoe(_Rhoe);
}

void ViscSolver::SetxxExt(size_t iSec, size_t iRegion, std::vector<double> _xxExt)
{   
    if (_xxExt.size() != Sections[iSec][iRegion]->mesh->GetnMarkers())
        throw std::runtime_error("dart::ViscSolver External mesh vector is not coherent with the mesh.");
    Sections[iSec][iRegion]->mesh->SetExt(_xxExt);
}

void ViscSolver::SetDeltaStarExt(size_t iSec, size_t iRegion, std::vector<double> _DeltaStarExt)
{
    if (_DeltaStarExt.size() != Sections[iSec][iRegion]->mesh->GetnMarkers())
    {
        std::cout << "size DeltaStar: " << _DeltaStarExt.size() << ". nMarkers: " << Sections[iSec][iRegion]->mesh->GetnMarkers() << std::endl;
        throw std::runtime_error("dart::ViscSolver External delta star vector is not coherent with the mesh.");
    }
    Sections[iSec][iRegion]->invBnd->SetDeltaStar(_DeltaStarExt);
}

std::vector<double> ViscSolver::ExtractxCoord(size_t iSec, size_t iRegion) const
{
    std::vector<double> xCoord(Sections[iSec][iRegion]->mesh->GetnMarkers(), 0.);
    for (size_t iPoint=0; iPoint<xCoord.size(); ++iPoint)
        xCoord[iPoint] = Sections[iSec][iRegion]->mesh->Getx(iPoint);
    return xCoord;
}

std::vector<double> ViscSolver::ExtractBlowingVel(size_t iSec, size_t iRegion) const
{
    return Sections[iSec][iRegion]->Blowing;
}
std::vector<double> ViscSolver::Extractxx(size_t iSec, size_t iRegion) const
{   
    std::vector<double> xx(Sections[iSec][iRegion]->mesh->GetnMarkers(),0.);
    for (size_t iPoint=0; iPoint<xx.size(); ++iPoint)
        xx[iPoint] = Sections[iSec][iRegion]->mesh->GetLoc(iPoint);
    return xx;
}
std::vector<double> ViscSolver::ExtractDeltaStar(size_t iSec, size_t iRegion) const
{
    return Sections[iSec][iRegion]->DeltaStar;
}

std::vector<double> ViscSolver::ExtractUe(size_t iSec, size_t iRegion) const
{
    std::vector<double> ueCurr(Sections[iSec][iRegion]->mesh->GetnMarkers(), 0.0);
    for (size_t i=0; i<ueCurr.size(); ++i)
        ueCurr[i] = Sections[iSec][iRegion]->U[i * Sections[iSec][iRegion]->GetnVar() + 3];
    return ueCurr;
}

void ViscSolver::SetUeOld(size_t iSec, size_t iRegion, std::vector<double> _ue)
{
    if (_ue.size() != Sections[iSec][iRegion]->mesh->GetnMarkers())
        throw std::runtime_error("dart::ViscSolver External velocity vector is not coherent with the mesh.");
    Sections[iSec][iRegion]->invBnd->SetVtOld(_ue);
}

std::vector<std::vector<double>> ViscSolver::GetSolution(size_t iSec)
{
    size_t nMarkersUp = Sections[iSec][0]->mesh->GetnMarkers();
    size_t nMarkersLw = Sections[iSec][1]->mesh->GetnMarkers() - 1; /* No stagnation point doublon */
    unsigned int nVar = Sections[iSec][0]->GetnVar();

    std::vector<std::vector<double>> Sln(16);
    for (size_t iVector=0; iVector<Sln.size(); ++iVector)
        Sln[iVector].resize(nMarkersUp + nMarkersLw + Sections[iSec][2]->mesh->GetnMarkers(), 0.);
    /* Upper side */
    size_t revPoint = nMarkersUp - 1;
    for (size_t iPoint=0; iPoint<nMarkersUp; ++iPoint)
    {
        Sln[0][iPoint] = Sections[iSec][0]->U[revPoint*nVar + 0];
        Sln[1][iPoint] = Sections[iSec][0]->U[revPoint*nVar + 1];
        Sln[2][iPoint] = Sections[iSec][0]->U[revPoint*nVar + 2];
        Sln[3][iPoint] = Sections[iSec][0]->U[revPoint*nVar + 3];
        Sln[4][iPoint] = Sections[iSec][0]->U[revPoint*nVar + 4];
        Sln[5][iPoint] = Sections[iSec][0]->DeltaStar[revPoint];
        
        Sln[6][iPoint] = Sections[iSec][0]->Ret[revPoint];
        Sln[7][iPoint] = Sections[iSec][0]->Cd[revPoint];
        Sln[8][iPoint] = Sections[iSec][0]->Cf[revPoint];
        Sln[9][iPoint] = Sections[iSec][0]->Hstar[revPoint];
        Sln[10][iPoint] = Sections[iSec][0]->Hstar2[revPoint];
        Sln[11][iPoint] = Sections[iSec][0]->Hk[revPoint];
        Sln[12][iPoint] = Sections[iSec][0]->Cteq[revPoint];
        Sln[13][iPoint] = Sections[iSec][0]->Us[revPoint];
        Sln[14][iPoint] = Sections[iSec][0]->Delta[revPoint];

        Sln[15][iPoint] = Sections[iSec][0]->mesh->Getx(revPoint);
        --revPoint;
    }
    /* Lower side */
    for (size_t iPoint=0; iPoint<Sections[iSec][1]->mesh->GetnMarkers(); ++iPoint)
    {
        Sln[0][nMarkersUp + iPoint] = Sections[iSec][1]->U[iPoint*nVar + 0];
        Sln[1][nMarkersUp + iPoint] = Sections[iSec][1]->U[iPoint*nVar + 1];
        Sln[2][nMarkersUp + iPoint] = Sections[iSec][1]->U[iPoint*nVar + 2];
        Sln[3][nMarkersUp + iPoint] = Sections[iSec][1]->U[iPoint*nVar + 3];
        Sln[4][nMarkersUp + iPoint] = Sections[iSec][1]->U[iPoint*nVar + 4];
        Sln[5][nMarkersUp + iPoint] = Sections[iSec][1]->DeltaStar[iPoint];
        
        Sln[6][nMarkersUp + iPoint] = Sections[iSec][1]->Ret[iPoint];
        Sln[7][nMarkersUp + iPoint] = Sections[iSec][1]->Cd[iPoint];
        Sln[8][nMarkersUp + iPoint] = Sections[iSec][1]->Cf[iPoint];
        Sln[9][nMarkersUp + iPoint] = Sections[iSec][1]->Hstar[iPoint];
        Sln[10][nMarkersUp + iPoint] = Sections[iSec][1]->Hstar2[iPoint];
        Sln[11][nMarkersUp + iPoint] = Sections[iSec][1]->Hk[iPoint];
        Sln[12][nMarkersUp + iPoint] = Sections[iSec][1]->Cteq[iPoint];
        Sln[13][nMarkersUp + iPoint] = Sections[iSec][1]->Us[iPoint];
        Sln[14][nMarkersUp + iPoint] = Sections[iSec][1]->Delta[iPoint];

        Sln[15][nMarkersUp + iPoint] = Sections[iSec][1]->mesh->Getx(iPoint);

    }
    /* Wake */
    for (size_t iPoint=0; iPoint<Sections[iSec][2]->mesh->GetnMarkers(); ++iPoint)
    {
        Sln[0][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->U[iPoint*nVar + 0];
        Sln[1][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->U[iPoint*nVar + 1];
        Sln[2][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->U[iPoint*nVar + 2];
        Sln[3][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->U[iPoint*nVar + 3];
        Sln[4][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->U[iPoint*nVar + 4];
        Sln[5][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->DeltaStar[iPoint];
        
        Sln[6][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Ret[iPoint];
        Sln[7][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Cd[iPoint];
        Sln[8][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Cf[iPoint];
        Sln[9][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Hstar[iPoint];
        Sln[10][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Hstar2[iPoint];
        Sln[11][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Hk[iPoint];
        Sln[12][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Cteq[iPoint];
        Sln[13][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Us[iPoint];
        Sln[14][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->Delta[iPoint];

        Sln[15][nMarkersUp + nMarkersLw + iPoint] = Sections[iSec][2]->mesh->Getx(iPoint);
    }

    if (verbose>2) std::cout << "Solution structure sent." << std::endl;

    return Sln;

}


void ViscSolver::PrintBanner() const
{
    std::cout << "\n\n"
              << std::endl;
    std::cout << "# ---------------------------------------------------------------------------- #" << std::endl;
    std::cout << "#                                                                              #" << std::endl;
    std::cout << "#                                                   ///,        ////           #" << std::endl;
    std::cout << "#       Paul Dechamps                               \\  /,      /  >.           #" << std::endl;
    std::cout << "#       ULiege 2022                                  \\  /,   _/  /.            #" << std::endl;
    std::cout << "#                                                     \\_  /_/   /.             #" << std::endl;
    std::cout << "#                                                      \\__/_   <               #" << std::endl;
    std::cout << "#                                                      /<<< \\_\\_               #" << std::endl;
    std::cout << "#   ██╗  ██╗ ██████╗ ██████╗ ██╗   ██╗███████╗        /,)^>>_._ \\              #" << std::endl;
    std::cout << "#   ██║  ██║██╔═══██╗██╔══██╗██║   ██║██╔════╝        (/   \\ /\\\\               #" << std::endl;
    std::cout << "#   ███████║██║   ██║██████╔╝██║   ██║███████╗            // ````              #" << std::endl;
    std::cout << "#   ██╔══██║██║   ██║██╔══██╗██║   ██║╚════██║     ======((`=======            #" << std::endl;
    std::cout << "#   ██║  ██║╚██████╔╝██║  ██║╚██████╔╝███████║     ___   ___   _               #" << std::endl;
    std::cout << "#   ╚═╝  ╚═╝ ╚═════╝ ╚═╝  ╚═╝ ╚═════╝ ╚══════╝    / __| | _ ) | |              #" << std::endl;
    std::cout << "#                                                | (__  | _ \\ | |__            #" << std::endl;
    std::cout << "#  High-Reynolds Operations for Robust Unsteady   \\___| |___/ |____|           #" << std::endl;
    std::cout << "#    and Separating Coupled Boundary layers                                    #" << std::endl;
    std::cout << "#                                                                              #" << std::endl;
    std::cout << "# ---------------------------------------------------------------------------- #" << std::endl;
    std::cout << "\n\n"
              << std::endl;
}