#!/usr/bin/env python3 
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute lifting (linear or nonlinear) viscous flow around a NACA 0012
# Amaury Bilocq
#
# Test the viscous-inviscid interaction scheme
# Reference to the master's thesis: http://hdl.handle.net/2268/252195
# Reference test cases with Naca0012 (different from master's thesis):
# 1) Incompressible: Re = 1e7, M_inf = 0, alpha = 5°, msTE = 0.01, msLE = 0.01
#    -> nIt = 27, Cl = 0.55, Cd = 0.0058, xtrTop = 0.087, xtrBot = 0.741
#    -> msLe = 0.001, xtrTop = 0.061
# 2) Compressible: Re = 1e7, M_inf = 5, alpha = 5°, msTE = 0.01, msLE = 0.01
#    -> nIt = 33, Cl = 0.65, Cd = 0.0063, xtrTop = 0.057, xtrBot = 0.740
# 3) Separated: Re = 1e7, M_inf = 0, alpha = 12°, msTE = 0.01, msLE = 0.001
#    -> nIt = 43, Cl = 1.30 , Cd = 0.0108, xtrTop = 0.010, xtrBot = 1
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import dart.utils as invUtils
import dart.default as invDefault

import vii.pyVII.vCoupler as viiCoupler
import vii.pyVII.vUtils as viscUtils

import tbox
import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors
from matplotlib import pyplot as plt

import math

def main():
    print('Start')
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    Re = 1e7
    alpha = 2.*math.pi/180
    M_inf = 0.715

    CFL0 = 1
    
    # ========================================================================================

    c_ref = 1
    dim = 2

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()

    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1, 'msTe' : 0.01, 'msLe' : 0.001}
    msh, gmshWriter = invDefault.mesh(dim, 'models/n0012.geo', pars, ['field', 'airfoil', 'downstream'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, blw = invDefault.problem(msh, dim, alpha, 0., M_inf, c_ref, c_ref, 0.25, 0., 0., 'airfoil', te = 'te', vsc = True, dbc=True)
    tms['pre'].stop()

    # solve viscous problem

    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    
    iSolver = invDefault.newton(pbl)
    vSolver = viscUtils.initBL(Re, M_inf, CFL0, 1)

    Coupler = viiCoupler.Coupler(iSolver, vSolver)
    Coupler.CreateProblem(blw[0], blw[1])
    Coupler.Run()

    # extract Cp
    Cp = invUtils.extract(bnd.groups[0].tag.elems, iSolver.Cp)
    tboxU.write(Cp, 'Cp.dat', '%1.5e', ', ', 'x, y, z, Cp', '')

    print('')
    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('      Re        M    alpha       Cl       Cd      Cdp      Cdf       Cm')
    print('{0:6.1f}e6 {1:8.2f} {2:8.1f} {3:8.4f} {4:8.4f} {5:8.4f} {6:8.4f} {7:8.4f}'.format(Re/1e6, M_inf, alpha*180/math.pi, iSolver.Cl, vSolver.Cdt, vSolver.Cdp, vSolver.Cdf, iSolver.Cm))

    # Write results to file
    vSolution = viscUtils.GetSolution(0, vSolver)
    viscUtils.WriteFile(vSolution, vSolver.GetRe())

    print('')
    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)
    print('SOLVERS statistics')
    print(Coupler.tms)

    
    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if Re == 1e7 and M_inf == 0. and alpha == 2.*math.pi/180:
        tests.add(CTest('Cl', iSolver.Cl, 0.2208, 5e-3))
        tests.add(CTest('Cd', vSolution['Cdt'], 0.00531, 1e-3, forceabs=True))
        tests.add(CTest('Cdp', vSolution['Cdp'], 0.0009, 1e-3, forceabs=True))
        tests.add(CTest('xtr_top', vSolution['xtrT'], 0.20, 0.03, forceabs=True))
        tests.add(CTest('xtr_bot', vSolution['xtrB'], 0.50, 0.03, forceabs=True))
    if Re == 1e7 and M_inf == 0. and alpha == 5.*math.pi/180:
        tests.add(CTest('Cl', iSolver.Cl, 0.554, 5e-3))
        tests.add(CTest('Cd', vSolution['Cdt'], 0.007, 1e-3, forceabs=True))
        tests.add(CTest('Cdp', vSolution['Cdp'],0.003, 1e-3, forceabs=True))
        tests.add(CTest('xtr_top', vSolution['xtrT'], 0.06, 0.03, forceabs=True))
        tests.add(CTest('xtr_bot', vSolution['xtrB'], 0.74, 0.03, forceabs=True))
    else:
        raise Exception('Test not defined for this flow')

    tests.run()


    # eof
    print('')

if __name__ == "__main__":
    main()
