#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute lifting (linear or nonlinear) potential flow around a NACA 0012 rectangular wing
# Adrien Crovato
#
# Test the nonlinear shock-capturing capability and the Kutta condition
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import dart.default as floD
import dart.utils as dartU # or from dartflo.dart import utils as dartU
import tbox.utils as tboxU # or from dartflo.tbox import utils as tboxU
import tboxVtk
import numpy as np

import vii.pyVII.vUtils as viscU
import vii.pyVII.vCoupler2 as viscC
import vii.pyVII.vInterpolator as vInterpol

from matplotlib import pyplot as plt

import fwk
from fwk.testing import *
from fwk.coloring import ccolors

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    """Re = 11.72e6
    alpha = 3.06*math.pi/180
    M_inf = 0.8395"""
    Re = 1e7
    alpha = 3.*math.pi/180
    M_inf = 0.4
    dim = 3
    CFL0 = 1

    # define dimension and mesh size
    spn = 1.0 # wing span
    lgt = 3.0 # channel half length
    hgt = 3.0 # channel half height
    wdt = 3.0 # channel width
    S_ref = 1.*spn # reference area
    c_ref = 1 # reference chord
    fms = 1.0 # farfield mesh size
    nms = 0.02 # nearfield mesh size

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'spn' : spn, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msLe' : nms, 'msTe' : nms, 'msF' : fms}
    imsh = viscU.mesh('/home/paul/lab/dartflo/dart/models/n0012_3.geo', pars)
    vmsh = viscU.mesh('/home/paul/lab/dartflo/vii/models/n0012_3_visc.geo', pars)
    msh, gmshWriter = floD.mesh(dim, 'models/n0012_3.geo', pars, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    tms['msh'].stop()

    # set the problem and add fluid, initial/boundary conditions
    tms['pre'].start()
    # Replce tp = 'teTip' by tp = 'wakeTip' for oneraM6.geo
    pbl, _, wk, bnd, blw = floD.problem(msh, dim, alpha, 0., M_inf, S_ref, c_ref, 0., 0., 0., 'wing', tp = 'teTip', vsc=True)
    tms['pre'].stop()

    config = {'nDim': dim, 'Sections':[0.026, 0.103, 0.179, 0.256, 0.333, 0.41, 0.436, 0.513, 0.59, 0.667, 0.744, 0.821, 0.897, 0.949], 'Sym':[0.], 'span':spn, 'rbftype': 'linear', 'smoothing': 1e-6, 'degree': 0, 'neighbors': 10}
    
    # solve problem
    iSolver = floD.newton(pbl)

    # Run inviscid calculation
    mshWriter = tboxVtk.VtkExport(msh)
    iSolver.run()
    iSolver.save(mshWriter)
    dartU.writeSlices(msh.name, config['Sections'], 5)

    cpInv = [np.zeros((0,0)) for _ in range(len(config['Sections']))]
    for i in range(len(config['Sections'])):
        cpInv[i] = tboxU.read('n0012_3_slice_'+str(i)+'.dat')
        #tboxU.plot(cpInv[i][:,3], cpInv[i][:,4], 'x', 'Cp', 'title', True)

    # VII
    iSolverAPI = vInterpol.Interpolator(iSolver, blw, imsh, vmsh, config)
    vSolver = viscU.initBL(Re, M_inf, CFL0, len(config['Sections']), spn, 2)
    coupler = viscC.Coupler(iSolverAPI, vSolver)

    coupler.Run()


    # Write corrected Cp
    iSolver.save(mshWriter)
    dartU.writeSlices(msh.name, config['Sections'], 5)

    cpCorr = [np.zeros((0,0)) for _ in range(len(config['Sections']))]
    for i in range(len(config['Sections'])):
        cpCorr[i] = tboxU.read('n0012_3_slice_'+str(i)+'.dat')

    # Plot cp

    for i in range(len(config['Sections'])):
        plt.plot(cpCorr[i][:,3], cpCorr[i][:,4], '-')
        plt.plot(cpInv[i][:,3], cpInv[i][:,4], '--')
        plt.xlabel('$x/c$')
        plt.ylabel('$Cp$')
        plt.title('$y/b$ = {:.3f}'.format(config['Sections'][i]))
        plt.gca().invert_yaxis()
        plt.legend(['Corrected', 'Inviscid'])
        plt.show()

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)
    print(coupler.tms)

    # visualize solution
    iSolverAPI.GetCp()
    iSolverAPI.iSolver.save(gmshWriter)
    floD.initViewer(pbl)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if alpha == 3*math.pi/180 and M_inf == 0.3 and spn == 1.0:
        tests.add(CTest('iteration count', solver.nIt, 3, 1, forceabs=True))
        tests.add(CTest('CL', solver.Cl, 0.135, 5e-2))
        tests.add(CTest('CD', solver.Cd, 0.0062, 1e-2)) # Tranair (NF=0.0062, FF=0.0030), Panair 0.0035
        tests.add(CTest('CS', solver.Cs, 0.0121, 5e-2))
        tests.add(CTest('CM', solver.Cm, -0.0278, 1e-2))
    else:
        raise Exception('Test not defined for this flow')
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
